import typing
import gi

gi.require_version("Gtk", "3.0")
gi.require_version("DbusmenuGtk3", "0.4")

from gi.repository import Gdk, Gtk, DbusmenuGtk3

from dasbus.connection import SessionMessageBus
from dasbus.client.observer import DBusObserver


class Menu(object):
    def __init__(self, service_name, object_path, settings, event_box: Gtk.EventBox, item):
        self.service_name = service_name
        self.object_path = object_path
        self.settings = settings
        self.event_box = event_box
        self.item = item
        self.session_bus = SessionMessageBus()
        self.menu_widget: typing.Union[None, DbusmenuGtk3.Menu] = None

        self.distance_scrolled_x = 0
        self.distance_scrolled_y = 0

        self.event_box.connect("button-press-event", self.button_press_event_handler)
        self.event_box.add_events(Gdk.EventMask.SCROLL_MASK | Gdk.EventMask.SMOOTH_SCROLL_MASK)
        self.event_box.connect("scroll-event", self.scroll_event_handler)

        self.menu_observer = DBusObserver(
            message_bus=self.session_bus,
            service_name=self.service_name
        )
        self.menu_observer.service_available.connect(
            self.menu_available_handler
        )
        self.menu_observer.service_unavailable.connect(
            self.menu_unavailable_handler
        )
        self.menu_observer.connect_once_available()

    def __del__(self):
        self.menu_observer.disconnect()
        self.session_bus.disconnect()

    def menu_available_handler(self, _observer):
        """print(
            "Menu -> menu_available_handler: Connecting to menu over dbus:\n"
            "  service_name: {}\n"
            "  object_path: {}".format(
                self.service_name,
                self.object_path
            )
        )"""
        self.menu_widget = DbusmenuGtk3.Menu().new(
            dbus_name=self.service_name,
            dbus_object=self.object_path
        )
        self.menu_widget.show()

    def menu_unavailable_handler(self, _observer):
        self.event_box.disconnect_by_func(self.button_press_event_handler)

    def button_press_event_handler(self, _w, event: Gdk.EventButton):
        if (event.button == 1 and self.item.item_is_menu) or event.button == 3:
            if self.menu_widget is not None:
                self.menu_widget.popup_at_widget(
                    self.event_box,
                    Gdk.Gravity.SOUTH,
                    Gdk.Gravity.NORTH,
                    event
                )
            else:
                self.item.context_menu(event)
        elif event.button == 1:
            self.item.activate(event)
        elif event.button == 2:
            self.item.secondary_action(event)

    def scroll_event_handler(self, _w, event: Gdk.EventScroll):
        dx = 0
        dy = 0
        if event.direction == Gdk.ScrollDirection.UP:
            dy = -1
        elif event.direction == Gdk.ScrollDirection.DOWN:
            dy = 1
        elif event.direction == Gdk.ScrollDirection.LEFT:
            dx = -1
        elif event.direction == Gdk.ScrollDirection.RIGHT:
            dx = 1
        elif event.direction == Gdk.ScrollDirection.SMOOTH:
            self.distance_scrolled_x += event.delta_x
            self.distance_scrolled_y += event.delta_y

            if self.distance_scrolled_x > self.settings["smooth-scrolling-threshold"]:
                dx = max((self.distance_scrolled_x, 1.0))
            elif self.distance_scrolled_x < self.settings["smooth-scrolling-threshold"]:
                dx = min((self.distance_scrolled_x, -1.0))

            if self.distance_scrolled_y > self.settings["smooth-scrolling-threshold"]:
                dy = max((self.distance_scrolled_y, 1.0))
            elif self.distance_scrolled_y > self.settings["smooth-scrolling-threshold"]:
                dy = min((self.distance_scrolled_y, -1.0))

        if dx != 0:
            self.item.scroll(dx, "horizontal")
        if dy != 0:
            self.item.scroll(dy, "vertical")



#!/usr/bin/env python3

"""
nwg-shell helper script to preview system processes
Copyright (c) 2023-2024 Piotr Miller
e-mail: nwg.piotr@gmail.com
GitHub: https://github.com/nwg-piotr/nwg-panel
Project: https://nwg-piotr.github.io/nwg-shell
License: MIT
"""

import json
import os
import socket
import sys
from enum import Enum

import psutil
from i3ipc import Connection
import gi

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk, GLib

from nwg_panel.tools import get_config_dir, load_json, save_json, check_key, eprint

swaysock = os.getenv('SWAYSOCK')
his = os.getenv("HYPRLAND_INSTANCE_SIGNATURE")
niri_sock = os.getenv("NIRI_SOCKET")


class SortOrder(Enum):
    NONE = 0
    PID = 1
    PPID = 2
    NAME = 3
    USERNAME = 4
    CPU_PERCENT = 5
    MEMORY_PERCENT = 6


sort_order = SortOrder.PID

# We need to get_allocated_width of each one inside a function later
btn_pid, btn_ppid, btn_owner, btn_cpu, btn_mem, btn_name = None, None, None, None, None, None,


def hyprctl(cmd):
    # /tmp/hypr moved to $XDG_RUNTIME_DIR/hypr in #5788
    xdg_runtime_dir = os.getenv("XDG_RUNTIME_DIR")
    hypr_dir = f"{xdg_runtime_dir}/hypr" if xdg_runtime_dir and os.path.isdir(
        f"{xdg_runtime_dir}/hypr") else "/tmp/hypr"

    s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    s.connect(f"{hypr_dir}/{os.getenv('HYPRLAND_INSTANCE_SIGNATURE')}/.socket.sock")

    s.send(cmd.encode("utf-8"))
    output = s.recv(20480).decode('utf-8')
    s.close()

    return output


def niri_ipc(cmd, is_json=False):
    niri_sock = os.getenv("NIRI_SOCKET")
    client = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    client.connect(niri_sock)
    if not is_json:
        client.send(f'"{cmd}"\n'.encode("utf-8"))
    else:
        client.send(f'{cmd}\n'.encode("utf-8"))

    buffer = ""
    while True:
        chunk = client.recv(1024).decode('utf-8', errors='replace')
        if not chunk:
            break
        buffer += chunk
    try:
        reply = json.loads(buffer)
        key = next(iter(reply))
        return reply[key]

    except json.JSONDecodeError as e:
        print("Failed to decode JSON:", e)
        print("Buffer:", buffer)
        return None


if not swaysock and not his and not niri_sock:
    eprint("Neither sway nor hyprland socket detected, terminating.")
    sys.exit(1)

W_OWNER = 10
W_NAME = 24

# Fallback icon names dict: win_name -> icon_name
aliases = {
    "Gimp-2.10": "gimp",
    "nwg-panel-config": "nwg-panel"
}

settings = {}  # nwg-panel common settings
scrolled_window = None
grid = Gtk.Grid()
window_lbl = None

theme = Gtk.IconTheme.get_default()


def handle_keyboard(win, event):
    if event.type == Gdk.EventType.KEY_RELEASE and event.keyval == Gdk.KEY_Escape:
        win.destroy()


def terminate(btn, pid):
    print("Terminating {}".format(pid))
    try:
        os.kill(pid, 15)
    except Exception as e:
        eprint(e)

    list_processes()


def list_processes(once=False):
    tree = None
    clients = {}
    windows = {}
    if swaysock:
        tree = Connection().get_tree()
    elif his:
        output = hyprctl("j/clients")
        clients = json.loads(output)
    elif niri_sock:
        command = "Windows"
        windows = niri_ipc(json.dumps(command), is_json=True)["Windows"]

    processes = {}

    user = os.getenv('USER')
    for proc in psutil.process_iter(['pid', 'ppid', 'name', 'username', 'cpu_percent', 'memory_percent']):
        if proc.info['username'] == os.getenv('USER') or not settings["processes-own-only"]:
            processes[proc.info['pid']] = proc.info

    processes_list = []
    for pid in processes:
        item = {
            "pid": pid,
            "ppid": processes[pid]["ppid"],
            "name": processes[pid]["name"],
            "username": processes[pid]["username"],
            "cpu_percent": processes[pid]["cpu_percent"],
            "memory_percent": processes[pid]["memory_percent"]
        }
        processes_list.append(item)

    if sort_order == SortOrder.PID:
        sorted_list = processes_list  # they are already sorted by PID, no need to sort
    elif sort_order == SortOrder.PPID:
        sorted_list = sorted(processes_list, key=lambda d: d['ppid'])
    elif sort_order == SortOrder.NAME:
        sorted_list = sorted(processes_list, key=lambda d: d['name'].upper())
    elif sort_order == SortOrder.USERNAME:
        sorted_list = sorted(processes_list, key=lambda d: d['username'].upper())
    elif sort_order == SortOrder.CPU_PERCENT:
        sorted_list = sorted(processes_list, key=lambda d: d['cpu_percent'], reverse=True)
    elif sort_order == SortOrder.MEMORY_PERCENT:
        sorted_list = sorted(processes_list, key=lambda d: d['memory_percent'], reverse=True)
    else:
        sorted_list = processes_list

    # At first, we need to add grid to the scrolled window (as in former add_with_viewport).
    # In next iterations, we add the grid directly to already existing viewport, to avoid the scrolled window floating.
    if scrolled_window and scrolled_window.get_children():
        viewport = scrolled_window.get_children()[0]
    else:
        viewport = None

    global grid
    if grid:
        grid.destroy()

    grid = Gtk.Grid.new()
    grid.set_column_spacing(3)

    if viewport:
        viewport.add(grid)
    elif scrolled_window:
        scrolled_window.add(grid)

    idx = 1
    for item in sorted_list:
        cons = None
        mapped = {}
        pid = item['pid']
        if swaysock:
            cons = tree.find_by_pid(pid)
        elif his:
            for client in clients:
                if client["pid"] == pid and client["mapped"]:
                    mapped["pid"] = client["class"]
                    break
        elif niri_sock:
            for window in windows:
                if window["pid"] == pid:
                    mapped["pid"] = window["app_id"]
                    break

        if not cons or not settings["processes-background-only"]:
            lbl = Gtk.Label.new(str(pid))
            lbl.set_xalign(0)
            grid.attach(lbl, 1, idx, 1, 1)

            lbl = Gtk.Label.new(str(processes[pid]["ppid"]))
            lbl.set_xalign(0)
            grid.attach(lbl, 2, idx, 1, 1)

            owner = processes[pid]["username"]
            if len(owner) > W_OWNER - 1:
                owner = "{}…".format(owner[:W_OWNER - 2])
            lbl = Gtk.Label.new(owner)
            lbl.set_xalign(0)
            grid.attach(lbl, 3, idx, 1, 1)

            percent = processes[pid]["cpu_percent"]
            if percent == 0:
                lbl = Gtk.Label.new("{}%".format(str(percent)))
            else:
                lbl = Gtk.Label()
                lbl.set_markup("<b>{}%</b>".format(str(percent)))
            lbl.set_xalign(0)
            grid.attach(lbl, 4, idx, 1, 1)

            percent = processes[pid]["memory_percent"]
            if percent < 1:
                lbl = Gtk.Label.new("{}%".format(str(round(percent, 2))))
            else:
                lbl = Gtk.Label()
                lbl.set_markup("<b>{}%</b>".format(str(round(percent, 2))))
            lbl.set_xalign(0)
            grid.attach(lbl, 5, idx, 1, 1)

            win_name = ""
            if cons:
                if cons[0].app_id:
                    win_name = cons[0].app_id
                elif cons[0].window_class:
                    win_name = cons[0].window_class
                elif cons[0].name:
                    win_name = cons[0].name
                elif cons[0].window_title:
                    win_name = cons[0].window_title
            elif mapped:
                win_name = mapped["pid"]

            if win_name:
                lbl = Gtk.Label.new(" {}".format(win_name))
                lbl.set_xalign(0)
                grid.attach(lbl, 8, idx, 1, 1)

            name = processes[pid]["name"]
            if theme.lookup_icon(name, 16, Gtk.IconLookupFlags.FORCE_SYMBOLIC):
                img = Gtk.Image.new_from_icon_name(name, Gtk.IconSize.MENU)
                img.set_property("halign", Gtk.Align.END)
                grid.attach(img, 6, idx, 1, 1)
            # fallback icon name
            elif win_name and theme.lookup_icon(win_name, 16, Gtk.IconLookupFlags.FORCE_SYMBOLIC):
                img = Gtk.Image.new_from_icon_name(win_name, Gtk.IconSize.MENU)
                img.set_property("halign", Gtk.Align.END)
                grid.attach(img, 6, idx, 1, 1)
            elif win_name and win_name in aliases and theme.lookup_icon(aliases[win_name], 16,
                                                                        Gtk.IconLookupFlags.FORCE_SYMBOLIC):
                img = Gtk.Image.new_from_icon_name(aliases[win_name], Gtk.IconSize.MENU)
                img.set_property("halign", Gtk.Align.END)
                grid.attach(img, 6, idx, 1, 1)

            if len(name) > W_NAME:
                name = "{}…".format(name[:W_NAME - 1])
            lbl = Gtk.Label.new(name)
            lbl.set_width_chars(W_NAME)
            lbl.set_xalign(0)
            grid.attach(lbl, 7, idx, 1, 1)

            if processes[pid]["username"] == user:
                btn = Gtk.Button.new_from_icon_name("gtk-close", Gtk.IconSize.MENU)
                btn.set_property("name", "btn-kill")
                btn.set_property("halign", Gtk.Align.CENTER)
                btn.connect("clicked", terminate, pid)
                grid.attach(btn, 0, idx, 1, 1)

            idx += 1

    # placeholders to align column width with the button box on top
    img = Gtk.Image()
    grid.attach(img, 0, idx + 1, 2, 1)
    img.set_size_request(btn_pid.get_allocated_width(), 0)

    img = Gtk.Image()
    grid.attach(img, 2, idx + 1, 1, 1)
    img.set_size_request(btn_ppid.get_allocated_width(), 0)

    img = Gtk.Image()
    grid.attach(img, 3, idx + 1, 1, 1)
    img.set_size_request(btn_owner.get_allocated_width(), 0)

    img = Gtk.Image()
    grid.attach(img, 4, idx + 1, 1, 1)
    img.set_size_request(btn_cpu.get_allocated_width(), 0)

    img = Gtk.Image()
    grid.attach(img, 5, idx + 1, 1, 1)
    img.set_size_request(btn_mem.get_allocated_width(), 0)

    img = Gtk.Image()
    grid.attach(img, 6, idx + 1, 2, 1)
    img.set_size_request(btn_name.get_allocated_width(), 0)

    grid.show_all()

    if not once:
        return True


def set_sort_order(btn, order):
    global sort_order
    sort_order = order

    btn_pid.set_label("  PID   ")
    btn_ppid.set_label(" PPID  ")
    btn_owner.set_label(" Owner  ")
    btn_cpu.set_label(" CPU%  ")
    btn_mem.set_label(" Mem%  ")
    btn_name.set_label("          Name          ")

    if order == SortOrder.PID:
        btn_pid.set_label("  PID  <")
    if order == SortOrder.PPID:
        btn_ppid.set_label(" PPID <")
    if order == SortOrder.USERNAME:
        btn_owner.set_label(" Owner <")
    if order == SortOrder.CPU_PERCENT:
        btn_cpu.set_label(" CPU% <")
    if order == SortOrder.MEMORY_PERCENT:
        btn_mem.set_label(" Mem% <")
    if order == SortOrder.NAME:
        btn_name.set_label("          Name         <")

    list_processes()


def on_background_cb(check_button):
    settings["processes-background-only"] = check_button.get_active()
    save_json(settings, os.path.join(get_config_dir(), "common-settings.json"))
    if window_lbl:
        window_lbl.set_visible(not settings["processes-background-only"])

    list_processes()


def on_own_cb(check_button):
    settings["processes-own-only"] = check_button.get_active()
    save_json(settings, os.path.join(get_config_dir(), "common-settings.json"))

    list_processes()


def main():
    GLib.set_prgname('nwg-processes')
    global settings
    settings = load_json(os.path.join(get_config_dir(), "common-settings.json"))
    defaults = {
        "processes-background-only": False,
        "processes-own-only": True,
        "processes-interval-ms": 2000
    }
    for key in defaults:
        check_key(settings, key, defaults[key])
    if not swaysock:
        settings["processes-background-only"] = False

    win = Gtk.Window.new(Gtk.WindowType.TOPLEVEL)
    win.connect('destroy', Gtk.main_quit)
    win.connect("key-release-event", handle_keyboard)

    box = Gtk.Box.new(Gtk.Orientation.VERTICAL, 6)
    box.set_property("margin", 6)
    box.set_property("vexpand", True)
    wrapper = Gtk.Box.new(Gtk.Orientation.VERTICAL, 0)
    box.pack_start(wrapper, False, False, 0)
    hbox = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 3)
    wrapper.pack_start(hbox, True, True, 0)

    global btn_pid, btn_ppid, btn_owner, btn_cpu, btn_mem, btn_name

    btn_pid = Gtk.Button.new_with_label("  PID  <")
    btn_pid.connect("clicked", set_sort_order, SortOrder.PID)
    hbox.pack_start(btn_pid, False, False, 0)
    btn_pid.show()

    btn_ppid = Gtk.Button.new_with_label(" PPID  ")
    btn_ppid.connect("clicked", set_sort_order, SortOrder.PPID)
    hbox.pack_start(btn_ppid, False, False, 0)
    btn_ppid.show()

    btn_owner = Gtk.Button.new_with_label(" Owner  ")
    btn_owner.connect("clicked", set_sort_order, SortOrder.USERNAME)
    hbox.pack_start(btn_owner, False, False, 0)

    btn_cpu = Gtk.Button.new_with_label(" CPU%  ")
    btn_cpu.connect("clicked", set_sort_order, SortOrder.CPU_PERCENT)
    hbox.pack_start(btn_cpu, False, False, 0)

    btn_mem = Gtk.Button.new_with_label(" Mem%  ")
    btn_mem.connect("clicked", set_sort_order, SortOrder.MEMORY_PERCENT)
    hbox.pack_start(btn_mem, False, False, 0)

    btn_name = Gtk.Button.new_with_label("          Name          ")
    btn_name.connect("clicked", set_sort_order, SortOrder.NAME)
    hbox.pack_start(btn_name, False, False, 0)

    global window_lbl
    window_lbl = Gtk.Label.new(" Window")
    window_lbl.set_xalign(0)
    hbox.pack_start(window_lbl, False, False, 0)

    win.add(box)

    global scrolled_window
    scrolled_window = Gtk.ScrolledWindow.new(None, None)
    scrolled_window.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
    scrolled_window.set_propagate_natural_height(True)
    box.pack_start(scrolled_window, True, True, 0)

    dist = Gtk.Box.new(Gtk.Orientation.VERTICAL, 0)
    dist.set_property("vexpand", True)
    box.pack_start(dist, True, True, 0)

    hbox = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 0)
    hbox.set_property("margin", 6)
    box.pack_start(hbox, False, False, 0)

    img = Gtk.Image.new_from_icon_name("nwg-processes", Gtk.IconSize.LARGE_TOOLBAR)
    hbox.pack_start(img, False, False, 6)

    lbl = Gtk.Label()
    lbl.set_markup("<b>nwg-processes</b>")
    hbox.pack_start(lbl, False, False, 0)

    if swaysock:
        cb = Gtk.CheckButton.new_with_label("Background only")
        cb.set_tooltip_text("Processes that don't belong to the sway tree")
        cb.set_active(settings["processes-background-only"])
        cb.connect("toggled", on_background_cb)
        hbox.pack_start(cb, False, False, 6)

    cb = Gtk.CheckButton.new_with_label("{}'s only".format(os.getenv('USER')))
    cb.set_tooltip_text("Processes that belong to the current $USER")
    cb.set_active(settings["processes-own-only"])
    cb.connect("toggled", on_own_cb)
    hbox.pack_start(cb, False, False, 6)

    btn = Gtk.Button.new_with_label("Close")
    hbox.pack_end(btn, False, False, 0)
    btn.connect("clicked", Gtk.main_quit)

    screen = Gdk.Screen.get_default()
    provider = Gtk.CssProvider()
    style_context = Gtk.StyleContext()
    style_context.add_provider_for_screen(screen, provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)
    css = b""" #btn-kill { padding: 0; border: 0; margin: 0 }
        label { font-family: DejaVu Sans Mono } """
    provider.load_from_data(css)

    win.show_all()

    win.set_size_request(0, 500)

    list_processes()

    if settings["processes-interval-ms"] > 0:
        Gdk.threads_add_timeout(GLib.PRIORITY_DEFAULT_IDLE, settings["processes-interval-ms"], list_processes)
    else:
        GLib.timeout_add(1000, list_processes, True)

    Gtk.main()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import json

from gi.repository import Gtk, Gdk

from nwg_panel.tools import niri_ipc, update_image, update_image_fallback_desktop


class NiriTaskbar(Gtk.Box):
    def __init__(self, settings, position, outputs, workspaces, windows, focused_window, display_name="",
                 icons_path=""):
        defaults = {
            "name-max-len": 24,
            "icon-size": 16,
            "workspaces-spacing": 0,
            "client-padding": 0,
            "workspace-label": "idx",
            "show-app-icon": True,
            "show-app-name": True,
            "show-layout": True,
            "all-outputs": False,
            "angle": 0.0
        }
        for key in defaults:
            if key not in settings:
                settings[key] = defaults[key]
        self.settings = settings

        self.position = position
        self.display_name = display_name
        self.icons_path = icons_path

        self.outputs = None
        self.active_workspaces = []
        self.windows = None
        self.ws_nums = None
        self.workspaces = [None]
        self.focused_window = None

        Gtk.Box.__init__(self, orientation=Gtk.Orientation.HORIZONTAL, spacing=settings["workspaces-spacing"])
        if self.settings["angle"] != 0.0:
            self.set_orientation(Gtk.Orientation.VERTICAL)

        self.refresh(outputs, workspaces, windows, focused_window)

    def parse_outputs(self, outputs):
        self.outputs = outputs

    def parse_workspaces(self, ws):
        self.ws_nums = []
        self.workspaces = {}
        self.active_workspaces = []
        for item in ws:
            self.ws_nums.append(item["id"])
            self.workspaces[item["id"]] = item
            if item["is_active"]:
                self.active_workspaces.append(item["id"])
        self.ws_nums.sort()

    def refresh(self, outputs, workspaces, windows, focused_window):
        self.outputs = outputs
        self.parse_workspaces(workspaces)
        self.windows = windows
        self.focused_window = focused_window
        for item in self.get_children():
            item.destroy()
        self.build_box()

    def build_box(self):
        for ws_num in self.ws_nums:
            ws_box = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 0)
            if self.settings["angle"] != 0.0:
                ws_box.set_orientation(Gtk.Orientation.VERTICAL)
            self.pack_start(ws_box, False, False, 0)
            if self.workspaces[ws_num]["output"] == self.display_name or self.settings["all-outputs"]:
                eb = Gtk.EventBox()

                ws_box.pack_start(eb, False, False, 6)
                lbl = Gtk.Label()
                if ws_num in self.active_workspaces:
                    if self.workspaces[ws_num]["name"]:
                        lbl.set_markup("<u>{}</u>:".format(self.workspaces[ws_num]["name"]))
                    else:
                        lbl.set_markup("<u>{}</u>:".format(self.workspaces[ws_num][self.settings["workspace-label"]]))
                else:
                    if self.workspaces[ws_num]["name"]:
                        lbl.set_markup("{}:".format(self.workspaces[ws_num]["name"]))
                    else:
                        lbl.set_text("{}:".format(self.workspaces[ws_num][self.settings["workspace-label"]]))
                eb.add(lbl)
                win_box = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 0)
                ws_box.pack_start(win_box, False, False, 0)
                for window in self.windows:
                    if window["workspace_id"] == ws_num:
                        client_box = ClientBox(self.settings, window, self.position, self.icons_path)
                        if self.focused_window and window["id"] == self.focused_window["id"]:
                            client_box.box.set_property("name", "task-box-focused")
                        else:
                            client_box.box.set_property("name", "task-box")
                        win_box.pack_start(client_box, False, False, self.settings["client-padding"])

        self.show_all()


def on_enter_notify_event(widget, event):
    widget.set_state_flags(Gtk.StateFlags.DROP_ACTIVE, clear=False)
    widget.set_state_flags(Gtk.StateFlags.SELECTED, clear=False)


def on_leave_notify_event(widget, event):
    widget.unset_state_flags(Gtk.StateFlags.DROP_ACTIVE)
    widget.unset_state_flags(Gtk.StateFlags.SELECTED)


class ClientBox(Gtk.EventBox):
    def __init__(self, settings, window, position, icons_path):
        self.position = position
        self.settings = settings
        self.id = window["id"]
        self.pid = window["pid"]
        self.icons_path = icons_path
        Gtk.EventBox.__init__(self)
        self.box = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, spacing=0)
        if settings["angle"] != 0.0:
            self.box.set_orientation(Gtk.Orientation.VERTICAL)
        self.add(self.box)

        self.connect('enter-notify-event', on_enter_notify_event)
        self.connect('leave-notify-event', on_leave_notify_event)
        self.connect('button-release-event', self.on_click, window, self.box)

        image = None
        if settings["show-app-icon"]:
            name = window["app_id"]
            image = Gtk.Image()
            image.set_property("name", "task-box-icon")
            update_image_fallback_desktop(image, name, settings["image-size"], icons_path)
            self.box.pack_start(image, False, False, 4)

        name = window["title"][:settings["name-max-len"]]

        if settings["show-app-name"]:
            lbl = Gtk.Label()
            lbl.set_angle(self.settings["angle"])

            lbl.set_text(name)
            self.box.pack_start(lbl, False, False, 6)

            if name and image:
                image.set_tooltip_text(name)

        if settings["show-layout"]:
            if window["is_floating"]:
                img = Gtk.Image()
                update_image(img, "focus-windows", self.settings["image-size"], self.icons_path)
                self.box.pack_start(img, False, False, 0)

    def on_click(self, widget, event, client, popup_at_widget):
        if event.button == 1:
            command = {"Action": {"FocusWindow": {"id": client["id"]}}}
            niri_ipc(json.dumps(command), is_json=True)

        if event.button == 3:
            menu = self.context_menu(client)
            menu.show_all()
            if self.position == "bottom":
                menu.popup_at_widget(popup_at_widget, Gdk.Gravity.SOUTH, Gdk.Gravity.NORTH, None)
            else:
                menu.popup_at_widget(popup_at_widget, Gdk.Gravity.NORTH, Gdk.Gravity.SOUTH, None)

    def context_menu(self, client):
        menu = Gtk.Menu()
        menu.set_reserve_toggle_size(False)

        # Toggle floating
        hbox = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 0)
        img = Gtk.Image()
        update_image(img, "view-paged-symbolic", 16, self.icons_path)
        hbox.pack_start(img, True, True, 0)
        img = Gtk.Image()
        update_image(img, "view-dual-symbolic", 16, self.icons_path)
        hbox.pack_start(img, True, True, 0)
        item = Gtk.MenuItem()
        item.add(hbox)
        item.connect("activate", self.toggle_floating)
        item.set_tooltip_text("toggle-window-floating")
        menu.append(item)

        # Fullscreen
        hbox = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 0)
        img = Gtk.Image()
        update_image(img, "view-fullscreen", 16, self.icons_path)
        hbox.pack_start(img, True, True, 0)
        item = Gtk.MenuItem()
        item.add(hbox)
        item.connect("activate", self.fullscreen)
        item.set_tooltip_text("fullscreen-window")
        menu.append(item)

        # Up
        hbox = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 0)
        img = Gtk.Image()
        update_image(img, "arrow-up", 16, self.icons_path)
        hbox.pack_start(img, True, True, 0)
        item = Gtk.MenuItem()
        item.add(hbox)
        item.connect("activate", self.up_ws_up)
        item.set_tooltip_text("move-window-up-or-to-workspace-up")
        menu.append(item)

        # Left
        hbox = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 0)
        img = Gtk.Image()
        update_image(img, "arrow-left", 16, self.icons_path)
        hbox.pack_start(img, True, True, 0)
        item = Gtk.MenuItem()
        item.add(hbox)
        item.connect("activate", self.left)
        item.set_tooltip_text("move-column-left-or-to-monitor-left")
        menu.append(item)

        # Right
        hbox = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 0)
        img = Gtk.Image()
        update_image(img, "arrow-right", 16, self.icons_path)
        hbox.pack_start(img, True, True, 0)
        item = Gtk.MenuItem()
        item.add(hbox)
        item.connect("activate", self.right)
        item.set_tooltip_text("move-column-right-or-to-monitor-right")
        menu.append(item)

        # Down
        hbox = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 0)
        img = Gtk.Image()
        update_image(img, "arrow-down", 16, self.icons_path)
        hbox.pack_start(img, True, True, 0)
        item = Gtk.MenuItem()
        item.add(hbox)
        item.connect("activate", self.down_ws_down)
        item.set_tooltip_text("move-window-down-or-to-workspace-down")
        menu.append(item)

        # Close
        hbox = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 0)
        img = Gtk.Image()
        update_image(img, "window-close", 16, self.icons_path)
        hbox.pack_start(img, True, True, 0)
        item = Gtk.MenuItem()
        item.add(hbox)
        item.connect("activate", self.close)
        item.set_tooltip_text("close-window")
        menu.append(item)

        return menu

    def close(self, *args):
        command = {"Action": {"CloseWindow": {"id": self.id}}}
        niri_ipc(json.dumps(command), is_json=True)

    def toggle_floating(self, *args):
        command = {"Action": {"ToggleWindowFloating": {"id": self.id}}}
        niri_ipc(json.dumps(command), is_json=True)

    def fullscreen(self, *args):
        command = {"Action": {"FullscreenWindow": {"id": self.id}}}
        niri_ipc(json.dumps(command), is_json=True)

    def up_ws_up(self, args):
        command = {"Action": {"MoveWindowUpOrToWorkspaceUp": {"id": self.id}}}
        niri_ipc(json.dumps(command), is_json=True)

    def left(self, args):
        command = {"Action": {"MoveColumnLeftOrToMonitorLeft": {}}}
        niri_ipc(json.dumps(command), is_json=True)

    def right(self, args):
        command = {"Action": {"MoveColumnRightOrToMonitorRight": {}}}
        niri_ipc(json.dumps(command), is_json=True)

    def down_ws_down(self, args):
        command = {"Action": {"MoveWindowDownOrToWorkspaceDown": {"id": self.id}}}
        niri_ipc(json.dumps(command), is_json=True)



import os
import sys
import gi

gi.require_version("Gtk", "3.0")

from gi.repository import Gtk, Gdk, GLib, GdkPixbuf

from nwg_panel.tools import check_key, create_pixbuf
from .item import StatusNotifierItem
from .menu import Menu


def resize_pix_buf(image, pixbuf, icon_size):
    # [At least on non-HiDPI system], this value is always in 1 or 2 (output scaled down or not scaled at all
    # / output scaled up), globally, whether one or more displays are scaled - so it seems useless.
    # E.g. if we scale one output * 1.2, we have icons resized * 2 on all outputs. Let's turn it off.
    """
    scaled_icon_size = image.get_scale_factor() * icon_size
    if pixbuf.get_height() != scaled_icon_size:
        width = scaled_icon_size * pixbuf.get_width() / pixbuf.get_height()
        pixbuf = pixbuf.scale_simple(width, scaled_icon_size, GdkPixbuf.InterpType.BILINEAR)
    """
    factor = 1
    w, h = pixbuf.get_width(), pixbuf.get_height()
    if w != icon_size:
        factor = icon_size / w
    elif h != icon_size:
        factor = icon_size / h
    pixbuf = pixbuf.scale_simple(w * factor, h * factor, GdkPixbuf.InterpType.BILINEAR)

    surface = Gdk.cairo_surface_create_from_pixbuf(pixbuf,
                                                   image.get_scale_factor(),
                                                   image.get_window())
    image.set_from_surface(surface)


def load_icon(image, icon_name: str, icon_size, icon_path=""):
    icon_size *= image.get_scale_factor()
    pixbuf = create_pixbuf(icon_name, icon_size, icon_path)
    resize_pix_buf(image, pixbuf, icon_size)


def update_icon(image, item, icon_size, icon_path):
    if "IconThemePath" in item.properties:
        icon_path = item.properties["IconThemePath"]
    load_icon(image, item.properties["IconName"], icon_size, icon_path)


def update_icon_from_pixmap(image, item, icon_size):
    largest_data = []
    largest_width = 0
    largest_height = 0
    for width, height, data in item.properties["IconPixmap"]:
        if width * height > largest_width * largest_height:
            largest_data = data
            largest_width = width
            largest_height = height

    # ARGB -> RGBA
    rgba_data = []
    for i in range(0, len(largest_data), 4):
        rgba_data += largest_data[i + 1:i + 4] + [largest_data[i]]

    pixbuf = GdkPixbuf.Pixbuf.new_from_data(
        rgba_data,
        GdkPixbuf.Colorspace.RGB,
        True,
        8,
        largest_width,
        largest_height,
        4 * largest_width
    )
    icon_size *= image.get_scale_factor()
    resize_pix_buf(image, pixbuf, icon_size)


def update_tooltip(image, item):
    icon_name, icon_data, title, description = item.properties["ToolTip"] if "ToolTip" in item.properties else item.properties["Tooltip"]
    tooltip = title
    if description:
        tooltip = "<b>{}</b>\n{}".format(title, description)
    image.set_tooltip_markup(tooltip)


def update_status(event_box, item):
    if "Status" in item.properties:
        status = item.properties["Status"].lower()
        event_box.set_visible(status != "passive")
        event_box_style = event_box.get_style_context()
        for class_name in event_box_style.list_classes():
            event_box_style.remove_class(class_name)
        if status == "needsattention":
            event_box_style.add_class("needs-attention")
        event_box_style.add_class(status)


class Tray(Gtk.EventBox):
    def __init__(self, settings, panel_position, icons_path=""):
        self.menu = None
        self.settings = settings
        self.icons_path = icons_path
        Gtk.EventBox.__init__(self)

        check_key(settings, "icon-size", 16)
        check_key(settings, "root-css-name", "tray")
        check_key(settings, "inner-css-name", "inner-tray")
        check_key(settings, "smooth-scrolling-threshold", 0)
        check_key(settings, "new-left", False)

        self.set_property("name", settings["root-css-name"])

        self.icon_size = settings["icon-size"]

        self.box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
        if panel_position in ["left", "right"]:
            self.box.set_orientation(Gtk.Orientation.VERTICAL)
        self.box.set_property("name", settings["inner-css-name"])
        self.add(self.box)

        self.items = {}

    def add_item(self, item: StatusNotifierItem):
        # print("Tray -> add_item: {}".format(item.properties))
        full_service_name = "{}{}".format(item.service_name, item.object_path)
        if full_service_name not in self.items:
            event_box = Gtk.EventBox()
            image = Gtk.Image()

            if "IconName" in item.properties and len(item.properties['IconName']) > 0:
                update_icon(image, item, self.icon_size, self.icons_path)
            elif "IconPixmap" in item.properties and len(item.properties["IconPixmap"]) != 0:
                update_icon_from_pixmap(image, item, self.icon_size)

            if "Tooltip" in item.properties or "ToolTip" in item.properties:
                update_tooltip(image, item)
            elif "Title" in item.properties:
                image.set_tooltip_markup(item.properties["Title"])

            update_status(event_box, item)

            event_box.add(image)
            if not self.settings["new-left"]:
                self.box.pack_start(event_box, False, False, 6)
            else:
                self.box.pack_end(event_box, False, False, 6)
            self.box.show_all()

            if "Menu" in item.properties:
                self.menu = Menu(
                    service_name=item.service_name,
                    object_path=item.properties["Menu"],
                    settings=self.settings,
                    event_box=event_box,
                    item=item
                )

            self.items[full_service_name] = {
                "event_box": event_box,
                "image": image,
                "item": item
            }

    def update_item(self, item: StatusNotifierItem, changed_properties: list[str]):
        full_service_name = "{}{}".format(item.service_name, item.object_path)
        event_box = self.items[full_service_name]["event_box"]
        image = self.items[full_service_name]["image"]

        def prop_changed(prop):
            return prop in changed_properties and len(item.properties[prop]) > 0

        if prop_changed("IconThemePath") or prop_changed("IconName"):
            update_icon(image, item, self.icon_size, self.icons_path)
        elif prop_changed("IconPixmap"):
            update_icon_from_pixmap(image, item, self.icon_size)

        if "Tooltip" in changed_properties or "ToolTip" in changed_properties:
            update_tooltip(image, item)
        elif "Title" in changed_properties:
            image.set_tooltip_markup(item.properties["Title"])

        update_status(event_box, item)

        event_box.show_all()

    def remove_item(self, item: StatusNotifierItem):
        full_service_name = "{}{}".format(item.service_name, item.object_path)
        self.box.remove(self.items[full_service_name]["event_box"])
        self.items.pop(full_service_name)
        self.box.show_all()



import typing
import os

from dasbus.connection import SessionMessageBus
from dasbus.client.observer import DBusObserver
from dasbus.client.proxy import disconnect_proxy

from .watcher import WATCHER_SERVICE_NAME, WATCHER_OBJECT_PATH
from .tray import Tray
from .item import StatusNotifierItem

HOST_SERVICE_NAME_TEMPLATE = "org.kde.StatusNotifierHost-{}-{}"
HOST_OBJECT_PATH_TEMPLATE = "/StatusNotifierHost/{}"


def get_service_name_and_object_path(service: str) -> (str, str):
    index = service.find("/")
    if index != len(service):
        return service[0:index], service[index:]
    return service, "/StatusNotifierItem"


class StatusNotifierHostInterface(object):
    def __init__(self, host_id, trays: typing.List[Tray]):
        self.host_id = host_id
        self.trays = trays

        self._statusNotifierItems = []
        self.watcher_proxy = None
        self.session_bus = SessionMessageBus()

        self.host_service_name = HOST_SERVICE_NAME_TEMPLATE.format(os.getpid(), self.host_id)
        self.host_object_path = HOST_OBJECT_PATH_TEMPLATE.format(self.host_id)
        self.session_bus.register_service(self.host_service_name)

        self.watcher_service_observer = DBusObserver(
            message_bus=self.session_bus,
            service_name=WATCHER_SERVICE_NAME
        )
        self.watcher_service_observer.service_available.connect(
            self.watcher_available_handler
        )
        self.watcher_service_observer.service_unavailable.connect(
            self.watcher_unavailable_handler
        )
        self.watcher_service_observer.connect_once_available()

    def __del__(self):
        if self.watcher_proxy is not None:
            disconnect_proxy(self.watcher_proxy)
        self.watcher_service_observer.disconnect()
        self.session_bus.disconnect()

    def watcher_available_handler(self, _observer):
        # print("StatusNotifierHostInterface -> watcher_available_handler")
        self.watcher_proxy = self.session_bus.get_proxy(WATCHER_SERVICE_NAME, WATCHER_OBJECT_PATH)
        self.watcher_proxy.StatusNotifierItemRegistered.connect(self.item_registered_handler)
        self.watcher_proxy.StatusNotifierItemUnregistered.connect(self.item_unregistered_handler)
        self.watcher_proxy.RegisterStatusNotifierHost(self.host_object_path, callback=lambda _: None)

        # Add items registered before host available
        for item in self.watcher_proxy.RegisteredStatusNotifierItems:
            self.item_registered_handler(item)

    def watcher_unavailable_handler(self, _observer):
        # print("StatusNotifierHostInterface -> watcher_unavailable_handler")
        self._statusNotifierItems.clear()
        disconnect_proxy(self.watcher_proxy)
        self.watcher_proxy = None

    def item_registered_handler(self, full_service_service):
        """print(
            "StatusNotifierHostInterface -> item_registered_handler\n  full_service_name: {}".format(
                full_service_service
            )
        )"""
        service_name, object_path = get_service_name_and_object_path(full_service_service)
        if self.find_item(service_name, object_path) is None:
            item = StatusNotifierItem(service_name, object_path)
            item.set_on_loaded_callback(self.item_loaded_handler)
            item.set_on_updated_callback(self.item_updated_handler)
            self._statusNotifierItems.append(item)

    def item_unregistered_handler(self, full_service_service):
        """print(
            "StatusNotifierHostInterface -> item_unregistered_handler\n  full_service_name: {}".format(
                full_service_service
            )
        )"""
        service_name, object_path = get_service_name_and_object_path(full_service_service)
        item = self.find_item(service_name, object_path)
        if item is not None:
            self._statusNotifierItems.remove(item)
            for tray in self.trays:
                tray.remove_item(item)

    def find_item(self, service_name, object_path) -> typing.Union[StatusNotifierItem, None]:
        for item in self._statusNotifierItems:
            if item.service_name == service_name and item.object_path == object_path:
                return item
        else:
            return None

    def item_loaded_handler(self, item):
        for tray in self.trays:
            tray.add_item(item)

    def item_updated_handler(self, item, changed_properties):
        for tray in self.trays:
            tray.update_item(item, changed_properties)


def init(host_id, trays: typing.List[Tray]):
    _status_notifier_host_interface = StatusNotifierHostInterface(host_id, trays)



#!/usr/bin/env python3
from enum import Enum
import os.path
import threading
from urllib.parse import unquote, urlparse

import gi

gi.require_version('Playerctl', '2.0')

from gi.repository import GLib, Gtk, Gdk
from gi.repository import Playerctl as Ctl
import requests

from nwg_panel.tools import check_key, eprint, local_dir, update_image


class Playerctl(Gtk.EventBox):
    PlayerOps = Enum('PlayerOps', ['PLAY_PAUSE', 'NEXT', 'PREVIOUS'])

    def __init__(self, settings, voc, icons_path=""):
        self.settings = settings
        self.icons_path = icons_path
        Gtk.EventBox.__init__(self)
        check_key(settings, "interval", 1)
        check_key(settings, "label-css-name", "")
        check_key(settings, "button-css-name", "")
        check_key(settings, "icon-size", 16)
        check_key(settings, "buttons-position", "left")
        check_key(settings, "chars", 30)
        check_key(settings, "scroll", True)
        check_key(settings, "show-cover", True)
        check_key(settings, "cover-size", 24)
        check_key(settings, "angle", 0.0)
        check_key(settings, "button-css-name", "")

        self.voc = voc

        self.old_cover_url = ""
        self.old_media_info = ""

        self.player = None
        self.player_handler_ids = []

        self.num_players = 0
        self.player_idx = 0
        self.add_events(Gdk.EventMask.SCROLL_MASK)
        self.connect('scroll-event', self.on_scroll)

        self.build_box()
        self.subscribe()

        # Hide on start if no player presents
        def hide_self(_):
            if not self.player:
                self.hide()

        self.connect("realize", hide_self)

    def subscribe(self):
        # Must associate manager with self to increase its reference count
        self.manager = Ctl.PlayerManager()
        self.manager.connect('name-appeared', self.on_name_appeared)
        self.manager.connect('player-vanished', self.on_player_vanished)

        # Manage all players from old to new, so that the newest one comes
        # first in props.players
        for name in reversed(self.manager.props.player_names):
            self.manage_player_by_name(self.manager, name)

        self.num_players = len(self.manager.props.players)
        if self.num_players > 1:
            self.num_players_lbl.set_text(f" {self.player_idx + 1}/{self.num_players} ")
            self.num_players_lbl.set_tooltip_text(
                f"Player {self.player_idx + 1}/{self.num_players}, {self.voc['scroll-to-switch']}")
        else:
            self.num_players_lbl.set_text("")

        if len(self.manager.props.players) > 0:
            self.init_player(self.manager.props.players[self.player_idx])

    @staticmethod
    def manage_player_by_name(manager, name):
        player = Ctl.Player.new_from_name(name)
        manager.manage_player(player)

    def on_name_appeared(self, manager, name):
        self.subscribe()
        self.deinit_player()
        self.manage_player_by_name(manager, name)
        self.init_player(manager.props.players[self.player_idx])

    def on_player_vanished(self, manager, player):
        self.subscribe()
        # Non-active player vanished, do nothing
        if self.player and player.props.player_name != self.player.props.player_name:
            return

        # Active player vanished, populate another one if exists
        self.deinit_player()
        if len(manager.props.players) > 0:
            self.init_player(manager.props.players[self.player_idx])

    def init_player(self, player):
        self.player = player
        self.show()

        # connect signals
        self.player_handler_ids.append(
            player.connect('metadata', self.on_metadata))
        self.player_handler_ids.append(
            player.connect('playback-status', self.on_playback_status))

        # Manually set the initial state
        # self.on_metadata(player, player.props.metadata)
        self.on_metadata(player, [])

    def deinit_player(self):
        if self.player:
            for handler_id in self.player_handler_ids:
                self.player.disconnect(handler_id)
        self.player = None
        self.player_handler_ids.clear()
        self.hide()

    def on_playback_status(self, player, status):
        artist = player.get_artist()
        title = player.get_title()
        status_text = None

        if status == Ctl.PlaybackStatus.PLAYING:
            update_image(self.play_pause_btn.get_image(), "media-playback-pause-symbolic",
                         self.settings["icon-size"], icons_path=self.icons_path)
        else:
            update_image(self.play_pause_btn.get_image(), "media-playback-start-symbolic",
                         self.settings["icon-size"], icons_path=self.icons_path)
            if status == Ctl.PlaybackStatus.PAUSED:
                status_text = "paused"
            elif status == Ctl.PlaybackStatus.STOPPED:
                status_text = "stopped"

        # Filter out empty value when building info
        info = [x for x in (artist, title, status_text) if x]
        info = " - ".join(info)
        self.set_media_info(info)

    def on_metadata(self, player, metadata):
        try:
            cover_url = metadata["mpris:artUrl"]
        except:  # used to be on KeyError, but actual error is 'mpris:artUrl' for some reason (playerctl bug?)
            cover_url = ""

        if cover_url != self.old_cover_url:
            self.old_cover_url = cover_url
            self.update_cover_image(cover_url)

        self.on_playback_status(player, player.props.playback_status)

    def update_remote_cover(self, url):
        try:
            r = requests.get(url, allow_redirects=True)
            cover_path = os.path.join(local_dir(), "cover.jpg")
            with open(cover_path, 'wb') as f:
                f.write(r.content)
            cover_path = "file://" + cover_path
        except Exception as e:
            eprint("Couldn't update remote cover: {}".format(e))
            cover_path = ""
        GLib.idle_add(self.update_cover_image, cover_path)

    def update_cover_image(self, url):
        url = urlparse(url)
        path = unquote(url.path)

        if url.scheme.startswith("http"):
            threading.Thread(target=self.update_remote_cover(url.geturl()), daemon=True).start()
            return

        if url.scheme == "file" and path:
            try:
                update_image(self.cover_img, path, self.settings["cover-size"], fallback=False)
            except Exception as e:
                eprint("Error creating pixbuf: {}".format(e))
                path = ""

        if not path:
            update_image(self.cover_img, "music", self.settings["cover-size"], self.icons_path)

    def on_scroll(self, widget, event):
        if event.direction == Gdk.ScrollDirection.UP:
            if self.player_idx < self.num_players - 1:
                self.player_idx += 1
            else:
                self.player_idx = 0
        if event.direction == Gdk.ScrollDirection.DOWN:
            if self.player_idx > 0:
                self.player_idx -= 1
            else:
                self.player_idx = self.num_players - 1
        print(f"Switched to player {self.player_idx}")
        self.subscribe()

    def build_box(self):
        self.box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
        if self.settings["angle"] != 0.0:
            self.box.set_orientation(Gtk.Orientation.VERTICAL)
        self.box.set_property("name", "task-box")
        self.add(self.box)

        button_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
        if self.settings["angle"] != 0.0:
            button_box.set_orientation(Gtk.Orientation.VERTICAL)

        img = Gtk.Image()
        update_image(img, "media-skip-backward-symbolic", self.settings["icon-size"], icons_path=self.icons_path)
        btn = Gtk.Button()
        btn.set_image(img)
        if self.settings["button-css-name"]:
            btn.set_property("name", self.settings["button-css-name"])
        btn.connect("clicked", self.launch, self.PlayerOps.PREVIOUS)
        button_box.pack_start(btn, False, False, 1)

        self.play_pause_btn = Gtk.Button()
        if self.settings["button-css-name"]:
            self.play_pause_btn.set_property("name", self.settings["button-css-name"])
        img = Gtk.Image()
        update_image(img, "media-playback-start-symbolic", self.settings["icon-size"], icons_path=self.icons_path)
        self.play_pause_btn.set_image(img)
        self.play_pause_btn.connect("clicked", self.launch, self.PlayerOps.PLAY_PAUSE)
        button_box.pack_start(self.play_pause_btn, False, False, 1)

        img = Gtk.Image()
        update_image(img, "media-skip-forward-symbolic", self.settings["icon-size"], icons_path=self.icons_path)
        btn = Gtk.Button()
        btn.set_image(img)
        if self.settings["button-css-name"]:
            btn.set_property("name", self.settings["button-css-name"])
        btn.connect("clicked", self.launch, self.PlayerOps.NEXT)
        button_box.pack_start(btn, False, False, 1)

        self.num_players_lbl = Gtk.Label.new("")
        if self.settings["label-css-name"]:
            self.num_players_lbl.set_property("name", self.settings["label-css-name"])
        self.num_players_lbl.set_angle(self.settings["angle"])

        self.label = AutoScrollLabel(self.settings["scroll"],
                                     self.settings["chars"],
                                     self.settings["interval"])
        if self.settings["label-css-name"]:
            self.label.set_property("name", self.settings["label-css-name"])
        self.label.set_angle(self.settings["angle"])

        self.cover_img = Gtk.Image()
        update_image(self.cover_img, "music", self.settings["cover-size"], self.icons_path)

        if self.settings["buttons-position"] == "left":
            self.box.pack_start(button_box, False, False, 2)
            if self.settings["show-cover"]:
                self.box.pack_start(self.cover_img, False, False, 0)
            self.box.pack_start(self.num_players_lbl, False, False, 0)
            self.box.pack_start(self.label, False, False, 5)
        else:
            if self.settings["show-cover"]:
                self.box.pack_start(self.cover_img, False, False, 2)
            self.box.pack_start(self.num_players_lbl, False, False, 0)
            self.box.pack_start(self.label, False, False, 2)
            self.box.pack_start(button_box, False, False, 10)

    def launch(self, button, op):
        if not self.player:
            return

        props = self.player.props

        if op == self.PlayerOps.PLAY_PAUSE:
            status = props.playback_status
            if status == Ctl.PlaybackStatus.PLAYING:
                if not props.can_pause:
                    return
            else:
                if not props.can_play:
                    return
            self.player.play_pause()

        elif op == self.PlayerOps.PREVIOUS:
            if props.can_go_previous:
                self.player.previous()
        elif op == self.PlayerOps.NEXT:
            if props.can_go_next:
                self.player.next()

    def set_media_info(self, text):
        if self.old_media_info != text:
            self.old_media_info = text
            self.label.set_tooltip_text(text)
            self.label.set_text(text)


class AutoScrollLabel(Gtk.Label):
    def __init__(self, scroll, chars, interval):
        super().__init__()
        self.chars = chars
        self.interval = interval if scroll else 0

        self.output_start_idx = 0
        self.text = ""
        self.src = 0

    def set_text(self, text):
        self.text = text
        self.output_start_idx = 0
        super().set_text(text[:self.chars])

        if self.interval == 0 or len(text) <= self.chars:
            # Disable scroll
            if self.src > 0:
                GLib.Source.remove(self.src)
                self.src = 0
        else:
            # Enable scroll
            if self.src == 0:
                self.src = GLib.timeout_add_seconds(self.interval,
                                                    self.scroll_text,
                                                    priority=GLib.PRIORITY_LOW)

    def scroll_text(self):
        self.output_start_idx += 1
        if self.output_start_idx + self.chars > len(self.text):
            self.output_start_idx = 0
        super().set_text(
            self.text[self.output_start_idx:self.output_start_idx + self.chars])
        return True



import os
import sys

class_to_icon_cache = {}
name_to_icon_cache = {}
filename_to_icon_cache = {}


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)

def __get_app_dirs():
    desktop_dirs = []

    home = os.getenv("HOME")
    xdg_data_home = os.getenv("XDG_DATA_HOME")
    xdg_data_dirs = os.getenv("XDG_DATA_DIRS") if os.getenv("XDG_DATA_DIRS") else "/usr/local/share/:/usr/share/"

    if xdg_data_home:
        desktop_dirs.append(os.path.join(xdg_data_home, "applications"))
    else:
        if home:
            desktop_dirs.append(os.path.join(home, ".local/share/applications"))

    for d in xdg_data_dirs.split(":"):
        desktop_dirs.append(os.path.join(d, "applications"))

    # Add flatpak dirs if not found in XDG_DATA_DIRS
    flatpak_dirs = [os.path.join(home, ".local/share/flatpak/exports/share/applications"),
                    "/var/lib/flatpak/exports/share/applications"]
    for d in flatpak_dirs:
        if d not in desktop_dirs:
            desktop_dirs.append(d)

    return desktop_dirs


def __process_desktop_file(file_path):
    try:
        with open(file_path, "r", encoding="utf-8") as f:
            content = f.read()
    except UnicodeDecodeError:
        eprint(f"Warning: Invalid .desktop file '{file_path}'")
        return
    except OSError as e:
        eprint(f"Warning: Unable to read .desktop file '{file_path}': {e}")
        return

    icon_name = None
    app_names = []
    startup_wm_class = None
    for line in content.splitlines():
        if line.startswith("[") and line != "[Desktop Entry]":
            break
        if line.upper().startswith("ICON"):
            icon_name = line.split("=", 1)[1].strip()
        elif line.upper().startswith("NAME"):
            app_names.append(line.split("=", 1)[1].strip())
        elif line.upper().startswith("STARTUPWMCLASS"):
            startup_wm_class = line.split("=", 1)[1].strip()

    if not icon_name:
        return

    if startup_wm_class:
        if not class_to_icon_cache.get(startup_wm_class):
            class_to_icon_cache[startup_wm_class] = icon_name
        # elif class_to_icon_cache.get(startup_wm_class) != icon_name:
        #     eprint(f"Warning: Duplicate class name '{startup_wm_class}' found in cache.")

    for app_name in app_names:
        if not name_to_icon_cache.get(app_name):
            name_to_icon_cache[app_name] = icon_name
        # elif name_to_icon_cache.get(app_name) != icon_name:
        #     eprint(f"Warning: Duplicate app name '{app_name}' found in cache.")

    base_filename = os.path.basename(file_path).upper()
    if not filename_to_icon_cache.get(base_filename):
        filename_to_icon_cache[base_filename] = icon_name
    elif filename_to_icon_cache.get(base_filename) != icon_name:
        eprint(f"Warning: Duplicate .desktop file name '{base_filename}' found in cache.")


def __populate_caches():
    app_dirs = __get_app_dirs()
    seen_dirs = set()
    while app_dirs:
        d = app_dirs.pop(0)
        if d in seen_dirs:
            continue
        seen_dirs.add(d)
        if os.path.isdir(d):
            try:
                files = os.listdir(d)
            except OSError as e:
                eprint(f"Warning: Can't list files in directory '{d}': {e}")
                continue

            for file_name in files:
                file_path = os.path.realpath(os.path.join(d, file_name))
                if not os.path.exists(file_path):
                    continue
                if os.path.isdir(file_path):
                    app_dirs.insert(0, file_path)
                    continue
                __process_desktop_file(file_path)


def get_icon_name(app_name):
    if not app_name:
        return ""

    if not class_to_icon_cache and not name_to_icon_cache:
        __populate_caches()

    # Search priority: window class > app name > .desktop filename
    if app_name in class_to_icon_cache:
        return class_to_icon_cache[app_name]
    if app_name in name_to_icon_cache:
        return name_to_icon_cache[app_name]
    for filename, icon in filename_to_icon_cache.items():
        if app_name.upper() in filename:
            return icon

    # GIMP returns "app_id": null and for some reason "class": "Gimp-2.10" instead of just "gimp".
    # Until the GTK3 version is released, let's make an exception for GIMP.
    if "GIMP" in app_name.upper():
        return "gimp"

    # if all above fails, use the app_name as icon name
    return app_name



import os
from setuptools import setup, find_packages


def read(f_name):
    return open(os.path.join(os.path.dirname(__file__), f_name)).read()


setup(
    name='nwg-panel',
    version='0.10.15',
    description='GTK3-based panel for sway and Hyprland Wayland compositors',
    packages=find_packages(),
    include_package_data=True,
    package_data={
        "": ["config/*", "icons_dark/*", "icons_light/*", "icons_color/*", "langs/*", "executors/*", "local/*",
             "modules/sni_system_tray/org.kde.StatusNotifierItem.xml"]
    },
    url='https://github.com/nwg-piotr/nwg-panel',
    license='MIT',
    author='Piotr Miller',
    author_email='nwg.piotr@gmail.com',
    python_requires='>=3.4.0',
    install_requires=['pygobject'],
    entry_points={
        'gui_scripts': [
            'nwg-panel = nwg_panel.main:main',
            'nwg-panel-config = nwg_panel.config:main',
            'nwg-dwl-interface = nwg_panel.dwl_interface:main',
            'nwg-processes = nwg_panel.processes:main'
        ]
    }
)



#!/usr/bin/env python3

from gi.repository import Gtk, GLib

from nwg_panel.tools import check_key, update_image, hyprctl


class HyprlandSubmap(Gtk.Box):
    def __init__(self, settings, icons_path=""):
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
        self.settings = settings
        self.label = Gtk.Label.new("default")
        self.submap = "default"

        icon = Gtk.Image()

        defaults = {
            "show-default": False,
            "show-icon": True,
            "css-name": "executor-label",
            "icon-size": 16,
            "angle": 0.0
        }
        for key in defaults:
            check_key(settings, key, defaults[key])

        update_image(icon, "object-flip-vertical", self.settings["icon-size"], icons_path)

        if self.settings["css-name"]:
            self.set_property("name", self.settings["css-name"])
        else:
            self.set_property("name", "executor-label")

        if settings["angle"] != 0.0:
            self.set_orientation(Gtk.Orientation.VERTICAL)
            self.label.set_angle(settings["angle"])

        if settings["show-icon"]:
            self.pack_start(icon, False, False, 0)

        self.pack_start(self.label, False, False, 6)

        self.refresh()

        self.connect("show", self.refresh)

        if self.submap == "default" and not self.settings["show-default"]:
            GLib.idle_add(self.hide, priority=GLib.PRIORITY_HIGH)

    def refresh(self, *args):
        self.submap = hyprctl("submap").strip()
        self.label.set_text(self.submap)
        if self.submap != "default" or self.settings["show-default"]:
            self.show_all()
        else:
            self.hide()



#!/usr/bin/python3

import subprocess


def get_output(cmd):
    try:
        return subprocess.check_output(cmd, shell=True).decode("utf-8").strip().splitlines()
    except subprocess.CalledProcessError:
        return ""


o = get_output("rfkill list")
for line in o:
    if "Soft blocked: yes" in line or "Hard blocked: yes" in line:
        print("airplane-mode-symbolic")
        print("")
        break



#!/usr/bin/env python3
import os.path

from gi.repository import Gtk, GLib
from i3ipc import Event

from nwg_panel.tools import check_key, get_icon_name, update_image, temp_dir, save_json
import nwg_panel.common


class Scratchpad(Gtk.Box):
    def __init__(self, i3, tree, settings, output, icons_path=""):
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
        self.settings = settings
        self.i3 = i3
        self.tree = tree
        self.content = []
        self.icons_path = icons_path
        self.output = output

        self.cache_file = os.path.join(temp_dir(), "nwg-scratchpad")

        defaults = {
            "css-name": "",
            "icon-size": 16,
            "angle": 0.0,
            "single-output": False
        }
        for key in defaults:
            check_key(settings, key, defaults[key])

        if settings["angle"] != 0.0:
            self.set_orientation(Gtk.Orientation.VERTICAL)

        self.check_scratchpad(i3.get_tree())
        self.subscribe()

    def subscribe(self):
        self.i3.on(Event.WINDOW, self.on_i3ipc_event)

    def check_scratchpad(self, tree):
        content = []

        scratchpad = tree.find_named('__i3_scratch')
        leaves = scratchpad[0].floating_nodes

        for node in leaves:
            aid = node.app_id if node.app_id else node.window_class
            if aid:
                pid = node.pid
                if node.app_id:
                    icon = get_icon_name(node.app_id)
                elif node.window_class:
                    icon = get_icon_name(node.window_class)
                else:
                    icon = "icon-missing"

                item = {"aid": aid, "pid": pid, "icon": icon, "name": node.name, "con_id": node.id}
                content.append(item)

        if content != self.content:
            self.content = content
            GLib.idle_add(self.build_box, priority=GLib.PRIORITY_HIGH)

    def build_box(self):
        for item in self.get_children():
            item.destroy()

        if len(self.content) > 0 and self.settings["css-name"]:
            self.set_property("name", self.settings["css-name"])
        else:
            self.set_property("name", None)

        for item in self.content:
            if item["icon"]:
                # Items to be listed on every scratchpad instance.
                if not self.settings["single-output"] or not nwg_panel.common.scratchpad_cons or str(
                        item["con_id"]) not in nwg_panel.common.scratchpad_cons:
                    self.add_event_box(item)
                # Items, that got to the scratchpad by the SwayTaskbar context menu, to be listed per-output.
                elif str(item["con_id"]) in nwg_panel.common.scratchpad_cons:
                    if self.output == nwg_panel.common.scratchpad_cons[str(item["con_id"])]["output"]:
                        self.add_event_box(item)

        self.show_all()

    def add_event_box(self, item):
        eb = Gtk.EventBox()
        image = Gtk.Image()
        update_image(image, item["icon"], self.settings["icon-size"], self.icons_path)
        eb.add(image)
        eb.connect("button-release-event", self.on_button_release, item["pid"], item["con_id"])
        if item["name"]:
            eb.set_tooltip_text(item["name"])
        self.pack_start(eb, False, False, 3)

    def on_button_release(self, eb, e, pid, con_id):
        if str(con_id) in nwg_panel.common.scratchpad_cons:
            # If moved to scratchpad with the SwayTaskbar context menu, we have stored
            # the workspace number and floating state. Let's restore them.
            item = nwg_panel.common.scratchpad_cons[str(con_id)]

            if "workspace" in item:
                if item["workspace"]:
                    # move to original workspace
                    cmd = "[con_id=\"{}\"] move to workspace number {}".format(con_id, item["workspace"])
                    self.i3.command(cmd)

            if "floating_con" in item:
                # restore floating state
                if item["floating_con"]:
                    cmd = "[con_id=\"{}\"] floating enable".format(con_id)
                else:
                    cmd = "[con_id=\"{}\"] floating disable".format(con_id)
                self.i3.command(cmd)

            # focus restored con
            cmd = "[con_id=\"{}\"] focus".format(con_id)
            self.i3.command(cmd)

            # remove the item from scratchpad cache
            nwg_panel.common.scratchpad_cons.pop(str(con_id), None)
            save_json(nwg_panel.common.scratchpad_cons, self.cache_file)

        else:
            # If moved to scratchpad in another way, we have no info on WS number and floating state.
            # Let's just show the item.
            cmd = "[pid={}] scratchpad show".format(pid)
            self.i3.command(cmd)

    def on_i3ipc_event(self, i3conn, event):
        self.check_scratchpad(i3conn.get_tree())



#!/usr/bin/env python3
import os.path

from gi.repository import GLib

import subprocess
from datetime import datetime

from nwg_panel.tools import (check_key, eprint, local_dir, load_json, save_json, update_image, update_gtk_entry,
                             create_background_task, cmd_through_compositor)

import gi

gi.require_version('Gtk', '3.0')
gi.require_version('Gdk', '3.0')

from gi.repository import Gtk, Gdk, GtkLayerShell


class Clock(Gtk.EventBox):
    def __init__(self, settings, icons_path=""):
        self.reminder_img_updated = False
        self.path = ""
        self.cal = None
        self.note_box = None
        self.popup = None
        self.note_entry = None
        self.icons_path = icons_path

        self.settings = settings
        Gtk.EventBox.__init__(self)
        self.box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)

        self.add(self.box)
        self.label = Gtk.Label.new("")

        defaults = {"root-css-name": "root-clock",
                    "css-name": "clock",
                    "tooltip-text": "",
                    "tooltip-date-format": False,
                    "on-right-click": "",
                    "on-middle-click": "",
                    "on-scroll-up": "",
                    "on-scroll-down": "",
                    "interval": 1,
                    "angle": 0.0,
                    "calendar-path": "",
                    "calendar-css-name": "calendar-window",
                    "calendar-placement": "top",
                    "calendar-margin-horizontal": 0,
                    "calendar-margin-vertical": 0,
                    "calendar-icon-size": 24,
                    "calendar-interval": 60,
                    "calendar-on": True}

        for key in defaults:
            check_key(settings, key, defaults[key])

        self.reminder_img = Gtk.Image()

        self.calendar = {}

        self.set_property("name", settings["root-css-name"])
        self.label.set_property("name", settings["css-name"])

        self.label.set_angle(settings["angle"])

        if settings["tooltip-text"]:
            self.set_tooltip_text(settings["tooltip-text"])

        if "format" not in settings:
            self.settings["format"] = "%a, %d. %b  %H:%M:%S"

        self.connect('button-release-event', self.on_button_release)
        if self.settings["calendar-on"] or self.settings["on-middle-click"] or self.settings["on-right-click"] or \
                self.settings["on-scroll-up"] or self.settings["on-scroll-down"]:
            self.connect('enter-notify-event', self.on_enter_notify_event)
        self.connect('leave-notify-event', self.on_leave_notify_event)
        if settings["on-scroll-up"] or settings["on-scroll-down"]:
            self.add_events(Gdk.EventMask.SCROLL_MASK)
            self.connect('scroll-event', self.on_scroll)

        self.load_calendar()
        self.build_box()
        self.refresh()

    def update_widget(self, output, tooltip=""):
        self.label.set_text(output)
        if self.settings["tooltip-date-format"] and tooltip:
            self.set_tooltip_text(tooltip)

        return False

    def get_output(self):
        now = datetime.now()
        try:
            time = now.strftime(self.settings["format"])
            tooltip = now.strftime(self.settings["tooltip-text"]) if self.settings["tooltip-date-format"] else ""
            GLib.idle_add(self.update_widget, time, tooltip)
        except Exception as e:
            print(e)

        ymd = now.strftime("%Y#%m#%d").split("#")
        y = ymd[0]
        try:
            month = int(ymd[1]) - 1
            m = str(month)
        except:
            m = None
        d = ymd[2]
        if self.has_note(y, m, d):
            if not self.reminder_img_updated:
                update_image(self.reminder_img, "gtk-apply", self.settings["calendar-icon-size"], self.icons_path)
                self.reminder_img_updated = True
            self.reminder_img.set_visible(True)
        else:
            self.reminder_img.set_visible(False)

    def refresh(self):
        thread = create_background_task(self.get_output, self.settings["interval"])
        thread.start()

        thread = create_background_task(self.reload_calendar, self.settings["calendar-interval"])
        thread.start()

    def build_box(self):
        if self.settings["calendar-on"]:
            self.box.pack_start(self.reminder_img, False, False, 0)
        self.box.pack_start(self.label, False, False, 6)

    def on_enter_notify_event(self, widget, event):
        widget.set_state_flags(Gtk.StateFlags.DROP_ACTIVE, clear=False)
        widget.set_state_flags(Gtk.StateFlags.SELECTED, clear=False)

    def on_leave_notify_event(self, widget, event):
        widget.unset_state_flags(Gtk.StateFlags.DROP_ACTIVE)
        widget.unset_state_flags(Gtk.StateFlags.SELECTED)

    def on_button_release(self, widget, event):
        if event.button == 1:
            if self.settings["calendar-on"]:
                self.display_calendar_window()
        elif event.button == 2 and self.settings["on-middle-click"]:
            self.launch(self.settings["on-middle-click"])
        elif event.button == 3 and self.settings["on-right-click"]:
            self.launch(self.settings["on-right-click"])

    def on_scroll(self, widget, event):
        if event.direction == Gdk.ScrollDirection.UP and self.settings["on-scroll-up"]:
            self.launch(self.settings["on-scroll-up"])
        elif event.direction == Gdk.ScrollDirection.DOWN and self.settings["on-scroll-up"]:
            self.launch(self.settings["on-scroll-down"])
        else:
            print("No command assigned")

    def launch(self, cmd):
        cmd = cmd_through_compositor(cmd)

        print(f"Executing: {cmd}")
        subprocess.Popen('{}'.format(cmd), shell=True)

    def display_calendar_window(self):
        if self.popup:
            if self.popup.is_visible():
                self.popup.destroy()
                return

            self.popup.destroy()

        self.load_calendar()

        self.popup = Gtk.Window.new(Gtk.WindowType.TOPLEVEL)
        self.popup.set_property("name", self.settings["calendar-css-name"])
        self.popup.connect("key-release-event", self.handle_keyboard)
        GtkLayerShell.init_for_window(self.popup)
        GtkLayerShell.set_layer(self.popup, GtkLayerShell.Layer.TOP)
        GtkLayerShell.set_keyboard_mode(self.popup, GtkLayerShell.KeyboardMode.ON_DEMAND)

        if self.settings["calendar-placement"] in ["top-left", "top", "top-right"]:
            GtkLayerShell.set_anchor(self.popup, GtkLayerShell.Edge.TOP, 1)
        elif self.settings["calendar-placement"] in ["bottom-left", "bottom", "bottom-right"]:
            GtkLayerShell.set_anchor(self.popup, GtkLayerShell.Edge.BOTTOM, 1)

        if self.settings["calendar-placement"] in ["top-left", "bottom-left"]:
            GtkLayerShell.set_anchor(self.popup, GtkLayerShell.Edge.LEFT, 1)
        elif self.settings["calendar-placement"] in ["top-right", "bottom-right"]:
            GtkLayerShell.set_anchor(self.popup, GtkLayerShell.Edge.RIGHT, 1)

        GtkLayerShell.set_margin(self.popup, GtkLayerShell.Edge.TOP, self.settings["calendar-margin-vertical"])
        GtkLayerShell.set_margin(self.popup, GtkLayerShell.Edge.BOTTOM, self.settings["calendar-margin-vertical"])
        GtkLayerShell.set_margin(self.popup, GtkLayerShell.Edge.RIGHT, self.settings["calendar-margin-horizontal"])
        GtkLayerShell.set_margin(self.popup, GtkLayerShell.Edge.LEFT, self.settings["calendar-margin-horizontal"])

        vbox = Gtk.Box.new(Gtk.Orientation.VERTICAL, 0)
        vbox.set_property("margin", 6)
        self.popup.add(vbox)
        self.cal = Gtk.Calendar.new()
        self.cal.connect("day-selected", self.on_day_selected)
        self.cal.connect("next-month", self.mark_days)
        self.cal.connect("prev-month", self.mark_days)
        self.cal.connect("next-year", self.mark_days)
        self.cal.connect("prev-year", self.mark_days)
        vbox.pack_start(self.cal, False, False, 0)

        self.popup.show_all()
        self.mark_days()

        self.note_box = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 0)
        self.note_entry = Gtk.Entry()
        self.note_entry.set_property("margin-top", 6)
        self.note_entry.connect("changed", self.on_note_changed)
        self.note_entry.connect("icon-release", self.on_note_icon_click)
        update_gtk_entry(self.note_entry, Gtk.EntryIconPosition.SECONDARY,
                         "edit-clear", self.settings["calendar-icon-size"],
                         self.icons_path)

        vbox.pack_start(self.note_box, False, False, 0)
        self.note_box.pack_start(self.note_entry, True, True, 0)
        btn = Gtk.Button()
        btn.set_property("margin-top", 6)
        img = Gtk.Image()
        btn.set_image(img)
        update_image(img, "gtk-close", self.settings["calendar-icon-size"], self.icons_path)
        btn.set_tooltip_text("Cancel & close")
        btn.connect("clicked", self.cancel_close_popup)
        btn.set_always_show_image(True)
        self.note_box.pack_start(btn, False, False, 0)

        btn = Gtk.Button()
        btn.set_property("margin-top", 6)
        img = Gtk.Image()
        btn.set_image(img)
        update_image(img, "object-select", self.settings["calendar-icon-size"], self.icons_path)
        btn.set_tooltip_text("Save & close")
        btn.connect("clicked", self.apply_close_popup)
        btn.set_always_show_image(True)
        self.note_box.pack_start(btn, False, False, 0)

        self.popup.set_size_request(self.popup.get_allocated_width() * 2, 0)

        y, m, d = self.cal.get_date()
        if self.has_note(y, m, d):
            self.cal.select_day(d)

    def mark_days(self, *args):
        for i in range(1, 32):
            self.cal.unmark_day(i)
        y, m, d = self.cal.get_date()
        y, m, d = str(y), str(m), str(d)
        if y in self.calendar and m in self.calendar[y]:
            for i in range(1, 32):
                if str(i) in self.calendar[y][m]:
                    self.cal.mark_day(i)

    def has_note(self, year, month, day):
        y = str(year)
        m = str(month)
        d = str(day)
        if y in self.calendar and m in self.calendar[y] and d in self.calendar[y][m] and self.calendar[y][m][d]:
            return True

        return False

    def cancel_close_popup(self, *args):
        self.popup.destroy()

    def apply_close_popup(self, *args):
        c = {}
        if len(self.calendar) > 0:
            for key_year in self.calendar:
                if len(self.calendar[key_year]) > 0:
                    for key_month in self.calendar[key_year]:
                        if len(self.calendar[key_year][key_month]) > 0:
                            for key_day in self.calendar[key_year][key_month]:
                                if self.calendar[key_year][key_month][key_day]:
                                    if key_year not in c:
                                        c[key_year] = {}
                                    if key_month not in c[key_year]:
                                        c[key_year][key_month] = {}
                                    note = self.calendar[key_year][key_month][key_day]
                                    if note:
                                        c[key_year][key_month][key_day] = note
        save_json(c, self.path)
        self.popup.destroy()

    def load_calendar(self):
        if self.settings["calendar-path"]:
            self.path = self.settings["calendar-path"]
            c = load_json(self.path)
            if c is not None:
                self.calendar = c
                return True
            else:
                result = save_json(self.calendar, self.path)
                if result == "ok":
                    print("Created new calendar file at '{}'".format(self.path))
                    return True
                else:
                    eprint("Couldn't create '{}': {}. Using default path.".format(self.path, result))

        self.path = os.path.join(local_dir(), "calendar.json")
        if not os.path.isfile(self.path):
            result = save_json(self.calendar, self.path)
            if result == "ok":
                print("Created new calendar file at '{}'".format(self.path))
            else:
                eprint("Couldn't create '{}': {}. No more idea...".format(self.path, result))
        c = load_json(self.path)
        self.calendar = c

    def reload_calendar(self):
        if not self.popup or not self.popup.is_visible():
            self.load_calendar()

    def handle_keyboard(self, win, event):
        if event.type == Gdk.EventType.KEY_RELEASE and event.keyval == Gdk.KEY_Escape:
            self.popup.destroy()

    def on_day_selected(self, cal):
        y, m, d = cal.get_date()
        y, m, d = str(y), str(m), str(d)
        if y in self.calendar and m in self.calendar[y] and d in self.calendar[y][m]:
            self.note_entry.set_text(self.calendar[y][m][d])
        else:
            self.note_entry.set_text("")

        self.note_entry.set_icon_sensitive(Gtk.EntryIconPosition.SECONDARY, self.note_entry.get_text())
        self.note_box.show_all()

    def on_note_changed(self, eb):
        y, m, d = self.cal.get_date()
        y, m, d = str(y), str(m), str(d)
        if y not in self.calendar:
            self.calendar[y] = {}
        if m not in self.calendar[y]:
            self.calendar[y][m] = {}
        note = eb.get_text()
        if note:
            self.note_entry.set_icon_sensitive(Gtk.EntryIconPosition.SECONDARY, True)
            self.calendar[y][m][d] = eb.get_text()
        elif d in self.calendar[y][m]:
            self.note_entry.set_icon_sensitive(Gtk.EntryIconPosition.SECONDARY, False)
            self.calendar[y][m].pop(d, None)

    def on_note_icon_click(self, entry, icon, event):
        entry.set_text("")
